from .section_labels import *

GEO_SECTIONS = [COORDINATES, VERTICES, POLYGONS]
GUI_SECTIONS = [MAP, SYMBOLS, LABELS, BACKDROP, PROFILES]
LINK_SECTIONS = [CONDUITS, ORIFICES, OUTLETS, PUMPS, WEIRS]
NODE_SECTIONS = [DIVIDERS, JUNCTIONS, OUTFALLS, STORAGE]
SUBCATCHMENT_SECTIONS = [SUBCATCHMENTS, SUBAREAS, INFILTRATION, POLYGONS]
POLLUTANT_SECTIONS = [TREATMENT, WASHOFF, BUILDUP, POLLUTANTS]
